import os
from typing import Any, Dict, Generator, List, Optional, Tuple
import json
from rdflib import Graph, Literal, Namespace, RDF, URIRef
from rdflib.term import BNode
from rdflib.namespace import SDO, XSD, PROV
import glob
from pyshacl import validate
from pathlib import Path

DSP = Namespace('http://ns.dasch.swiss/repository#')


# Conversion Methods
# ------------------


def convert_and_save(files: List[str], target: str) -> int:
    """Convert a list of metadata files and save output to files.

    Takes a list of `.json` files, converts each of them, and stores it to the specified target directory.

    Args:
        files (List[str]): paths to `.json` files that should be transformed.
        target (str): path to the target directory where the output files should be saved.

    Returns:
        int: The number of files that were converted.
    """
    res = 0
    for f in files:
        tup = convert_file(f)
        if not tup:
            continue
        ttl, jsonld = tup
        print(f'File length: {len(ttl)}')
        fpath = Path(f)
        out_ttl = f"{target}/{fpath.stem}.ttl"
        out_jsonld = f"{target}/{fpath.stem}.jsonld"
        with open(out_ttl, mode='w+', encoding='utf-8') as fw:
            fw.write(ttl)
        with open(out_jsonld, mode='w+', encoding='utf-8') as fw:
            fw.write(jsonld)
        res += 1
    return res


def convert_string(data_string: str, filename: str) -> Graph:
    """Convert the string representation of a V2 JSON metadata set to an RDF graph.

    Args:
        data_string (str): A string of V2 JSON metadata
        filename (str): the filename of the input for error logging purposes

    Returns:
        Graph: a Graph representing the RDF mapping of the input metadata
    """
    data: Dict[str, Any] = json.loads(data_string)
    g = Graph()
    g.bind('dsp', DSP)
    g.bind('rdf', RDF)
    g.bind('sdo', SDO)
    g.bind('prov', PROV)

    project = data.get('project')
    if not project:
        raise Exception(f'Metadata set {filename} did not contain a project, which is required.')
    _get_project(g, project)

    datasets = data.get('datasets')
    if not datasets:
        raise Exception(f'Metadata set "{filename}.json" did not contain any datasets, at least one is required.')
    for dataset in datasets:
        _get_dataset(g, dataset)

    persons = data.get('persons')
    if persons:
        for p in persons:
            _get_person(g, p)

    organizations = data.get('organizations')
    if organizations:
        for o in organizations:
            _get_organization(g, o)

    grants = data.get('grants')
    if grants:
        for gr in grants:
            _get_grant(g, gr)

    return g


def convert_file(file: str) -> Optional[Tuple[str, str]]:
    """Convert a file containing V2 JSON metadata set to RDF serializations.

    Args:
        file (str): path of a JSON file

    Returns:
        Optional[Tuple[str, str]]: None if the conversion failed. Otherwise a tuple with the turtle and the JSON-LD serialization of the metadata.
    """
    filename = file.rsplit('.', 1)[0]
    try:
        with open(file, mode='r+', encoding='utf-8') as f:
            data = f.read()
        graph = convert_string(data, filename)
        ttl = graph.serialize()
        ttl = ttl.replace('<http://schema.org/>', '<https://schema.org/>')
        jsonld = graph.serialize(format='json-ld')
        jsonld = jsonld.replace('<http://schema.org/>', '<https://schema.org/>')
        return ttl, jsonld
    except Exception as e:
        print(f'WARNING: failed to convert file {file}! ({type(e).__name__} caused by: {e})')
        return None


def convert_files(files: List[str]) -> None:
    """Convert multiple files"""
    for file in files:
        convert_file(file)


# RDF Mapping Methods
# -------------------


def _get_project(g: Graph, d: Dict[str, Any]) -> None:
    iri = URIRef(d['__id'])
    g.add((iri, RDF.type, DSP.Project))

    prop = d['shortcode']
    g.add((iri, DSP.hasShortcode, Literal(prop, datatype=XSD.string)))

    prop = d['name']
    g.add((iri, DSP.hasName, Literal(prop, datatype=XSD.string)))

    prop = d['howToCite']
    g.add((iri, DSP.hasHowToCite, Literal(prop, datatype=XSD.string)))

    prop = d['description']
    for txt in __get_multi_language_text(g, prop):
        g.add((iri, DSP.hasDescription, txt))

    prop = d['startDate']
    g.add((iri, DSP.hasStartDate, Literal(prop, datatype=XSD.date)))

    prop = d['teaserText']
    g.add((iri, DSP.hasTeaser, Literal(prop, datatype=XSD.string)))

    props = d['datasets']
    for prop in props:
        g.add((iri, DSP.hasDataset, URIRef(prop)))

    props = d['keywords']
    for prop in props:
        for txt in __get_multi_language_text(g, prop):
            g.add((iri, DSP.hasKeyword, txt))

    props = d['disciplines']
    for prop in props:
        if prop.get('__type') == 'URL':
            bn_url = __get_url(g, prop)
            g.add((iri, DSP.hasDiscipline, bn_url))
        else:
            for txt in __get_multi_language_text(g, prop):
                g.add((iri, DSP.hasDiscipline, txt))

    props = d['temporalCoverage']
    for prop in props:
        if prop.get('__type') == 'URL':
            bn_url = __get_url(g, prop)
            g.add((iri, DSP.hasTemporalCoverage, bn_url))
        else:
            for txt in __get_multi_language_text(g, prop):
                g.add((iri, DSP.hasTemporalCoverage, txt))

    props = d['spatialCoverage']
    for prop in props:
        bn_url = __get_url(g, prop)
        g.add((iri, DSP.hasSpatialCoverage, bn_url))

    props = d['funders']
    for prop in props:
        g.add((iri, DSP.hasFunder, URIRef(prop)))

    prop = d['url']
    bn_url = __get_url(g, prop)
    g.add((iri, DSP.hasURL, bn_url))

    prop = d.get('secondaryURL')
    if prop:
        bn_url = __get_url(g, prop)
        g.add((iri, DSP.hasSecondaryURL, bn_url))

    prop = d.get('dataManagementPlan')
    if prop:
        dmp = BNode()
        av = prop.get('available')
        if av:
            g.add((dmp, DSP.isAvailable, Literal(True, datatype=XSD.boolean)))
        url = prop.get('url')
        if url:
            bn_url = __get_url(g, prop)
            g.add((dmp, DSP.hasURL, bn_url))
        if av or url:
            g.add((dmp, RDF.type, DSP.DataManagementPlan))
            g.add((iri, DSP.hasDataManagementPlan, dmp))

    prop = d.get('endDate')
    if prop:
        g.add((iri, DSP.hasEndDate, Literal(prop, datatype=XSD.date)))

    prop = d.get('contactPoint')
    if prop:
        g.add((iri, DSP.hasContactPoint, URIRef(prop)))

    props = d.get('publications')
    if props:
        for prop in props:
            g.add((iri, DSP.hasPublication, Literal(prop, datatype=XSD.string)))

    props = d.get('grants')
    if props:
        for prop in props:
            g.add((iri, DSP.hasGrant, URIRef(prop)))

    props = d.get('alternativeNames')
    if props:
        for prop in props:
            for txt in __get_multi_language_text(g, prop):
                g.add((iri, DSP.hasAlternativeName, txt))


def _get_dataset(g: Graph, d: Dict[str, Any]) -> None:
    iri = URIRef(d['__id'])
    g.add((iri, RDF.type, DSP.Dataset))

    title = d['title']
    g.add((iri, DSP.hasTitle, Literal(title, datatype=XSD.string)))

    conds = d['accessConditions']
    g.add((iri, DSP.hasAccessConditions, Literal(conds, datatype=XSD.string)))

    htc = d['howToCite']
    g.add((iri, DSP.hasHowToCite, Literal(htc, datatype=XSD.string)))

    status = d['status']
    g.add((iri, DSP.hasStatus, Literal(status, datatype=XSD.string)))

    abstracts = d['abstracts']
    for abstract in abstracts:
        if abstract.get('__type') == 'URL':
            bn_url = __get_url(g, abstract)
            g.add((iri, DSP.hasAbstract, bn_url))
        else:
            for txt in __get_multi_language_text(g, abstract):
                g.add((iri, DSP.hasAbstract, txt))

    types = d['typeOfData']
    for t in types:
        g.add((iri, DSP.hasTypeOfData, Literal(t, datatype=XSD.string)))

    licenses = d['licenses']
    for lic in licenses:
        license = BNode()
        g.add((license, RDF.type, DSP.License))
        bn_url = __get_url(g, lic['license'])
        g.add((license, DSP.hasURL, bn_url))
        g.add((license, DSP.hasDate, Literal(lic['date'], datatype=XSD.date)))
        if lic.get('details'):
            g.add((license, DSP.hasDetails, Literal(lic['details'], datatype=XSD.string)))
        g.add((iri, DSP.hasLicense, license))

    langs = d['languages']
    for lang in langs:
        for txt in __get_multi_language_text(g, lang):
            g.add((iri, DSP.hasLanguage, txt))

    attrs = d['attributions']
    for attr in attrs:
        attribution = BNode()
        g.add((attribution, RDF.type, PROV.Attribution))
        g.add((attribution, PROV.agent, URIRef(attr['agent'])))
        for role in attr['roles']:
            g.add((attribution, DSP.hasRole, Literal(role, datatype=XSD.string)))
        g.add((iri, DSP.hasQualifiedAttribution, attribution))

    date = d.get('datePublished')
    if date:
        g.add((iri, DSP.hasDatePublished, Literal(date, datatype=XSD.date)))

    date = d.get('dateCreated')
    if date:
        g.add((iri, DSP.hasDateCreated, Literal(date, datatype=XSD.date)))

    date = d.get('dateModified')
    if date:
        g.add((iri, DSP.hasDateModified, Literal(date, datatype=XSD.date)))

    distro = d.get('distribution')
    if distro:
        bn_url = BNode()
        g.add((bn_url, RDF.type, SDO.DataDownload))
        g.add((bn_url, SDO.url, Literal(distro['url'])))
        g.add((iri, DSP.hasDistribution, bn_url))

    alt_titles = d.get('alternativeTitles')
    if alt_titles:
        for title in alt_titles:
            for txt in __get_multi_language_text(g, title):
                g.add((iri, DSP.hasAlternativeTitle, txt))

    urls = d.get('urls')
    if urls:
        for url in urls:
            bn_url = __get_url(g, url)
            g.add((iri, DSP.hasURL, bn_url))

    additional = d.get('additional')
    if additional:
        for add in additional:
            if abstract.get('__type') == 'URL':
                bn_url = __get_url(g, abstract)
                g.add((iri, DSP.hasAdditional, bn_url))
            else:
                for txt in __get_multi_language_text(g, abstract):
                    g.add((iri, DSP.hasAdditional, txt))


def _get_person(g: Graph, d: Dict[str, Any]) -> None:
    iri = URIRef(d['__id'])
    g.add((iri, RDF.type, DSP.Person))

    titles = d['jobTitles']
    for title in titles:
        g.add((iri, DSP.hasJobTitle, Literal(title, datatype=XSD.string)))

    given_names = d['givenNames']
    given_name = ' '.join(given_names)
    g.add((iri, DSP.hasGivenName, Literal(given_name, datatype=XSD.string)))

    family_names = d['familyNames']
    family_name = ' '.join(family_names)
    g.add((iri, DSP.hasFamilyName, Literal(family_name, datatype=XSD.string)))

    affiliations = d['affiliation']
    for affiliation in affiliations:
        g.add((iri, DSP.hasAffiliation, URIRef(affiliation)))

    addr = d.get('address')
    if addr:
        address = __get_address(g, addr)
        g.add((iri, DSP.hasAddress, address))

    email = d.get('email')
    if email:
        g.add((iri, DSP.hasEmail, Literal(email, datatype=XSD.string)))

    email = d.get('secondaryEmail')
    if email:
        g.add((iri, DSP.hasSecondaryEmail, Literal(email, datatype=XSD.string)))

    authRefs = d.get('authorityRefs')
    if authRefs:
        for ref in authRefs:
            bn_url = __get_url(g, ref)
            g.add((iri, DSP.hasAuthorityFileReference, bn_url))


def _get_organization(g: Graph, d: Dict[str, Any]) -> None:
    iri = URIRef(d['__id'])
    g.add((iri, RDF.type, DSP.Organization))

    name = d['name']
    g.add((iri, DSP.hasName, Literal(name, datatype=XSD.string)))

    url = d.get('url')
    if url:
        bn_url = __get_url(g, url)
        g.add((iri, DSP.hasURL, bn_url))

    addr = d.get('address')
    if addr:
        address = __get_address(g, addr)
        g.add((iri, DSP.hasAddress, address))

    email = d.get('email')
    if email:
        g.add((iri, DSP.hasEmail, Literal(email, datatype=XSD.string)))

    names = d.get('alternativeNames')
    if names:
        for n_multilang in names:
            for txt in __get_multi_language_text(g, n_multilang):
                g.add((iri, DSP.hasAbstract, txt))

    authRefs = d.get('authorityRefs')
    if authRefs:
        for ref in authRefs:
            bn_url = __get_url(g, ref)
            g.add((iri, DSP.hasAuthorityFileReference, bn_url))


def _get_grant(g: Graph, d: Dict[str, Any]) -> None:
    iri = URIRef(d['__id'])
    g.add((iri, RDF.type, DSP.Grant))

    funders = d['funders']
    for funder in funders:
        g.add((iri, DSP.hasFunder, URIRef(funder)))

    nr = d.get('number')
    if nr:
        g.add((iri, DSP.hasNumber, Literal(nr, datatype=XSD.string)))

    n = d.get('name')
    if n:
        g.add((iri, DSP.hasName, Literal(n, datatype=XSD.string)))

    url = d.get('url')
    if url:
        bn_url = __get_url(g, url)
        g.add((iri, DSP.hasURL, bn_url))


# Helper Methods
# --------------


def __get_url(g: Graph, d: Dict[str, Any]) -> BNode:
    text = d.get('text')
    url = d.get('url')
    bn_url = BNode()
    g.add((bn_url, RDF.type, SDO.URL))
    g.add((bn_url, SDO.url, Literal(url)))
    if text and not text == url:
        propID_bn = BNode()
        g.add((propID_bn, RDF.type, SDO.PropertyValue))
        g.add((propID_bn, SDO.propertyID, Literal(text)))
        g.add((bn_url, SDO.propertyID, propID_bn))
    return bn_url


def __get_address(g: Graph, d: Dict[str, Any]) -> BNode:
    bn_addr = BNode()
    g.add((bn_addr, RDF.type, SDO.PostalAddress))
    g.add((bn_addr, SDO.streetAddress, Literal(d['street'], datatype=XSD.string)))
    if d.get('additional'):
        g.add((bn_addr, DSP.additional, Literal(d['additional'], datatype=XSD.string)))
    g.add((bn_addr, SDO.postalCode, Literal(d['postalCode'], datatype=XSD.string)))
    g.add((bn_addr, SDO.addressLocality, Literal(d['locality'], datatype=XSD.string)))
    if d.get('canton'):
        g.add((bn_addr, SDO.addressRegion, Literal(d['canton'], datatype=XSD.string)))
    g.add((bn_addr, SDO.addressCountry, Literal(d['country'], datatype=XSD.string)))
    return bn_addr


def __get_multi_language_text(g: Graph, d: Dict[str, Any]) -> Generator[Literal, None, None]:
    for lang in d.keys():
        txt = d[lang]
        yield Literal(txt, lang=lang)


# Validation Functions
# --------------------


def validate_files(files: List[str]) -> None:
    """Validate multiple files"""
    for file in files:
        validate_file(file)


def validate_file(file: str) -> None:
    """Validate a V2 RDF metadata file against the SHACL ontology"""
    shacl = "https://raw.githubusercontent.com/dasch-swiss/dsp-meta-svc/main/docs/services/metadata/schema-metadata.shacl"
    print(file)
    conforms, result_graph, result_text = validate(data_graph=file,
                                                   shacl_graph=shacl,
                                                   inference='none')
    print(result_text)
    print('\n-------------\n\n')



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dsp-metadata-gui",
    version="1.2.0",
    description="Python GUI tool to collect metadata for DSP projects.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dasch-swiss/dsp-metadata-gui",
    author="Balduin Landolt",
    author_email="balduin.landolt@dasch.swiss",
    license="GPLv3",
    packages=["dspMetadataGUI", "dspMetadataGUI.util"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9.0",
    install_requires=[
        "decorator==4.4.2",
        "isodate==0.6.0",
        "numpy==1.20.1; python_version >= '3.0'",
        "owlrl==5.2.1",
        "pillow==8.1.0; python_version >= '3.6'",
        "pyparsing==2.4.7; python_version >= '2.6' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "pyshacl==0.14.2",
        "rdflib==5.0.0",
        "rdflib-jsonld==0.5.0",
        "six==1.15.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "validators==0.18.2",
        "wxpython==4.1.1",
    ],
    entry_points={
        "console_scripts": [
            "dsp-metadata=dspMetadataGUI.collectMetadata:collectMetadata"
        ],
    },
    include_package_data=True,
    zip_safe=False,
)



"""
Module to convert RDF serialized metadata (first data model) into JSON metadata (new data model).
"""

import re
import os
from pathlib import Path
from typing import Any, Dict, List
from rdflib import Graph
from rdflib.namespace import Namespace, RDF, SDO, PROV, SKOS
import json
import jsonschema
from rdflib.term import BNode, Literal
import requests
import time
from datetime import datetime
from textblob import TextBlob
import guess_language
import langid
from langdetect import detect
from bs4 import BeautifulSoup


schema_url = "https://raw.githubusercontent.com/dasch-swiss/dsp-meta-svc/main/docs/services/metadata/schema-metadata.json"
dsp = Namespace("http://ns.dasch.swiss/repository#")


def convert_and_save(files: List[str], target: str) -> int:
    """Convert a list of metadata files and save output to files.

    Takes a list of `.ttl` files, converts each of them, and stores it to the specified target directory.

    Args:
        files (List[str]): paths to `.ttl` files that should be transformed.
        target (str): path to the target directory where the output files should be saved.

    Returns:
        int: The number of files that were converted.
    """
    res = 0
    for f in files:
        s = convert_file(f)
        print(f'File length: {len(s)}')
        fpath = Path(f)
        out = f"{target}/{fpath.stem}.json"
        print(out)
        with open(out, mode='w+', encoding='utf-8') as fw:
            fw.write(s)
        res += 1
    return res


def convert_files(files: List[str]) -> List[str]:
    """Convert multiple metadata files from a list of files.

    Convert metadata from local .ttl files.

    Args:
        files (List[str]): paths to .ttl files (can be relative or absolute)

    Returns:
        List[str]: list of json serialized metadata
    """
    return [convert_file(f) for f in files]


def convert_file(file: str) -> str:
    """Convert metadata from a file.

    Convert metadata from a local .ttl file.

    Args:
        file (str): path to a .ttl file (can be relative or absolute)

    Returns:
        str: json serialized metadata
    """
    with open(file, 'r+', encoding='utf-8') as f:
        s = f.read()
    return convert_string(s)


def convert_strings(data_list: List[str]) -> List[str]:
    """Convert multiple metadata sets from a list of strings.

    Args:
        data_list (List[str]): list of turtle serializations of metadata

    Returns:
        List[str]: list of json serialized metadata
    """
    return [convert_string(data) for data in data_list]


def convert_string(data: str) -> str:
    """Convert metadata from a string.

    Args:
        data (str): string of a turtle setrialization of metadata

    Returns:
        str: json serialized metadata
    """
    data = data.replace('@prefix schema: <https://schema.org/> .', '@prefix schema: <http://schema.org/> .')
    g = Graph()
    g.parse(data=data, format='ttl')
    res = {"$schema": schema_url}
    res['project'] = _get_project(g)
    res['datasets'] = _get_datasets(g)
    for ds in res.get('datasets'):  # type: ignore
        res['project']['datasets'].append(ds.get('__id'))  # type: ignore
    persons = _get_persons(g)
    if persons:
        res['persons'] = persons
    orgs = _get_organizations(g)
    if orgs:
        res['organizations'] = orgs
    grants = _get_grants(g)
    if grants:
        res['grants'] = grants

    validate(res)
    return json.dumps(res, indent=4, ensure_ascii=False)


# DMP
# ---


def _get_dmp(g: Graph):
    """Get data management plan from graph"""
    try:
        dmp = next(g.subjects(RDF.type, dsp.DataManagementPlan))
    except StopIteration:
        return {}

    res: Dict[str, Any] = {"__type": "DataManagementPlan"}

    for _, p, o in g.triples((dmp, None, None)):
        obj = str(o)
        if p == dsp.hasURL:
            res['url'] = _get_url(g, o)
        elif p == dsp.isAvailable:
            res['available'] = obj == "true"
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Grant:", p, obj)

    return res


# Grant
# -----

def _get_grants(g: Graph):
    """Get grants from graph"""
    grants = g.subjects(RDF.type, dsp.Grant)
    return [_get_grant(g, grant) for grant in grants]


def _get_grant(g: Graph, grant_iri) -> Dict[str, Any]:
    """Get grant from graph"""
    res = {"__id": grant_iri,
           "__type": "Grant",
           "__createdAt": str(time.time_ns()),
           "__createdBy": "dsp-metadata-gui", }

    for _, p, o in g.triples((grant_iri, None, None)):
        obj = str(o).strip()
        if p == dsp.hasFunder:
            res.setdefault('funders', [])
            res['funders'].append(obj)
        elif p == dsp.hasNumber:
            res['number'] = obj
        elif p == dsp.hasName:
            res['name'] = obj
        elif p == dsp.hasURL:
            res['url'] = _get_url(g, o)
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Grant:", p, obj)

    return res


# Organizations
# -------------

def _get_organizations(g: Graph):
    """Get organizations from graph"""
    orgs = g.subjects(RDF.type, dsp.Organization)
    return [_get_organization(g, org) for org in orgs]


def _get_organization(g: Graph, org_iri):
    """Get organization from graph"""
    res = {"__id": org_iri,
           "__type": "Organization",
           "__createdAt": str(time.time_ns()),
           "__createdBy": "dsp-metadata-gui", }

    for _, p, o in g.triples((org_iri, None, None)):
        obj = str(o).strip()
        if p == dsp.hasName:
            res['name'] = obj
        elif p == dsp.hasURL:
            res['url'] = _get_url(g, o)
        elif p == dsp.hasEmail:
            res['email'] = obj
        elif p == dsp.hasAddress:
            res['address'] = _get_address(g, o)
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Organization:", p, obj)

    return res


# Person
# ------

def _get_persons(g: Graph):
    """Get persons from graph"""
    persons = g.subjects(RDF.type, dsp.Person)
    return [_get_person(g, person) for person in persons]


def _get_person(g: Graph, person_iri):
    """Get person from graph"""
    res = {"__id": person_iri,
           "__type": "Person",
           "__createdAt": str(time.time_ns()),
           "__createdBy": "dsp-metadata-gui", }

    for _, p, o in g.triples((person_iri, None, None)):
        obj = str(o).strip()
        if p == dsp.hasJobTitle:
            res.setdefault('jobTitles', [])
            res['jobTitles'].append(obj)
        elif p == dsp.hasGivenName:
            res['givenNames'] = obj.split(';')
        elif p == dsp.hasFamilyName:
            res['familyNames'] = obj.split(';')
        elif p == dsp.isMemberOf:
            res.setdefault('affiliation', [])
            res['affiliation'].append(obj)
        elif p == dsp.hasAddress:
            res['address'] = _get_address(g, o)
        elif p == dsp.hasEmail:
            if res.get('email'):
                res['secondaryEmail'] = obj
            else:
                res['email'] = obj
        elif p == dsp.sameAs:
            res.setdefault('authorityRefs', [])
            res['authorityRefs'].append(_get_url(g, o))
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Person:", p, obj)

    return res


# Dataset
# -------

def _get_datasets(g: Graph):
    """Get datasets from graph"""
    datasets = g.subjects(RDF.type, dsp.Dataset)
    return [_get_dataset(g, dataset) for dataset in datasets]


def _get_dataset(g: Graph, dataset_iri):
    """Get dataset from graph"""
    res = {"__id": dataset_iri,
           "__type": "Dataset",
           "__createdAt": str(time.time_ns()),
           "__createdBy": "dsp-metadata-gui", }

    for _, p, o in g.triples((dataset_iri, None, None)):
        obj = str(o).strip()
        if p == dsp.hasTitle:
            res['title'] = obj
        elif p == dsp.hasConditionsOfAccess:
            obj = obj.lower()
            if obj not in ['open', 'restricted', 'closed']:
                obj = f"XX - Access Conditions should be one of 'open', 'restricted' or 'closed', but found: {obj}"
            res['accessConditions'] = obj
        elif p == dsp.hasHowToCite:
            res['howToCite'] = obj
        elif p == dsp.hasStatus:
            res['status'] = obj
        elif p == dsp.hasAbstract:
            res.setdefault("abstracts", [])
            if isinstance(o, BNode):
                res['abstracts'].append(_get_url(g, o))
            else:
                res['abstracts'].append(_guess_language_of_text(obj))
        elif p == dsp.hasTypeOfData:
            res.setdefault("typeOfData", [])
            if obj == "Movie":
                obj = "Video"
            res['typeOfData'].append(obj)
        elif p == dsp.hasLicense:
            res.setdefault("licenses", [])
            res['licenses'].append({"__type": "License",
                                    "date": str(datetime.date(datetime.now())),
                                    "license": _get_url(g, o)})
        elif p == dsp.hasLanguage:
            res.setdefault("languages", [])
            res['languages'].append(_get_language(obj))
        elif p == dsp.hasQualifiedAttribution:
            res.setdefault("attributions", [])
            res['attributions'] = (_add_attribution(g, o, res['attributions']))
        elif p == dsp.hasAlternativeTitle:
            res.setdefault("alternativeTitles", [])
            res['alternativeTitles'].append(_guess_language_of_text(obj))
        elif p == dsp.hasDateCreated:
            res['dateCreated'] = obj
        elif p == dsp.hasDatePublished:
            res['datePublished'] = obj
        elif p == dsp.hasDateModified:
            res['dateModified'] = obj
        elif p == dsp.hasDistribution:
            res['distribution'] = _get_url(g, o)
        elif p == dsp.sameAs:
            res.setdefault('urls', [])
            res['urls'].append(_get_url(g, o))
        elif p == dsp.hasDocumentation:
            res.setdefault("additional", [])
            if isinstance(o, BNode):
                res['additional'].append(_get_url(g, o))
            else:
                res['additional'].append(_guess_language_of_text(obj))
        elif p == dsp.isPartOf:
            pass
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Dataset:", p, obj)

    return res


# Project
# -------

def _get_project(g: Graph):
    """Get project from graph"""
    project_iri = next(g.triples((None, RDF.type, dsp.Project)))[0]
    res = {"__id": project_iri,
           "__type": "Project",
           "__createdAt": str(time.time_ns()),
           "__createdBy": "dsp-metadata-gui",
           "howToCite": "XX - new property on project level",
           "teaserText": "XX - new property",
           "datasets": []}

    for _, p, o in g.triples((project_iri, None, None)):
        obj = str(o).strip()
        if p == dsp.hasShortcode:
            res['shortcode'] = obj
        elif p == dsp.hasName:
            res['name'] = obj
        elif p == dsp.hasDescription:
            res['description'] = _guess_language_of_text(obj)
        elif p == dsp.hasStartDate:
            res['startDate'] = obj
        elif p == dsp.hasEndDate:
            res['endDate'] = obj
        elif p == dsp.hasKeywords:
            res.setdefault('keywords', [])
            res['keywords'].append(_guess_language_of_text(obj))
        elif p == dsp.hasDiscipline:
            res.setdefault('disciplines', [])
            if isinstance(o, BNode):
                res['disciplines'].append(_get_url(g, o))
            else:
                res['disciplines'].append(_guess_language_of_text(obj))
        elif p == dsp.hasTemporalCoverage:
            res.setdefault('temporalCoverage', [])
            if isinstance(o, BNode):
                res['temporalCoverage'].append(_get_url(g, o))
            else:
                res['temporalCoverage'].append(_guess_language_of_text(obj))
        elif p == dsp.hasSpatialCoverage:
            res.setdefault('spatialCoverage', [])
            res['spatialCoverage'].append(_get_place(g, o))
        elif p == dsp.hasURL:
            if res.get('url'):
                res['secondaryURL'] = _get_url(g, o)
            else:
                res['url'] = _get_url(g, o)
        elif p == dsp.hasDataManagementPlan:
            res['dataManagementPlan'] = _get_dmp(g)
        elif p == dsp.hasPublication:
            res.setdefault('publications', [])
            res['publications'].append(obj)
        elif p == dsp.hasGrant:
            res.setdefault('grants', [])
            res['grants'].append(obj)
        elif p == dsp.hasAlternateName:
            res.setdefault('alternativeNames', [])
            res['alternativeNames'].append(_guess_language_of_text(obj))
        elif p == dsp.hasFunder:
            res.setdefault('funders', [])
            res['funders'].append(obj)
        elif p == dsp.hasContactPoint:
            res['contactPoint'] = obj
        # default cases
        elif p == RDF.type:
            pass
        else:
            print("Issue: Could not handle in Project:", p, obj)

    return res


# Utils
# -----

def _add_attribution(g: Graph, iri: BNode, attributions: List):
    """From a graph g, add an attribution defined by a blank node (i.e. all tripples that have the iri as subject), to a list."""
    role = str(next(g.objects(iri, dsp.hasRole)))
    agent = str(next(g.objects(iri, PROV.agent)))
    for att in attributions:
        p = att.get('agent')
        r = att.get('roles')
        if r and p and p == agent:
            r.append(role)
            return attributions
    attributions.append({"__type": "Attribution",
                         "agent": agent,
                         "roles": [role]})
    return attributions


def _get_language(language):
    """Get the correct language representation according to a string describing a language (i.e. the language itself or a language shortcode)"""
    language = str(language)
    if re.match('^[a-z]{2}$', language):
        return _get_language_from_shortcode(language)
    else:
        return _guess_language_of_text(language)


def _get_language_from_shortcode(code):
    """Get language representation from a language shortcode"""
    code = str(code)
    langs = {'en': {'en': 'English',
                    'de': 'Englisch',
                    'fr': 'anglais'},
             'de': {'en': 'German',
                    'de': 'Deutsch',
                    'fr': 'allemand'},
             'fr': {'en': 'French',
                    'de': 'Französisch',
                    'fr': 'français'}}
    default = {code: 'XX - ' + code}
    lang = langs.get(code)
    return lang if lang else default


def _guess_language_of_text(text):
    """ Guess the language of a string.

    Returns a dict of type multilanguage text, if possible with the correct language attribution to the text.
    """
    probables = ["en", "de", "fr"]
    text = str(text)
    predefined = _get_predefined_multilanguage_string(text)
    if predefined:
        return predefined
    try:
        lang_txtblob = TextBlob(text).detect_language()
        if lang_txtblob in probables:
            return {lang_txtblob: text}
    except Exception:
        lang_txtblob = None
    try:
        guess = guess_language.guess_language(text, probables)
        if guess in probables:
            return {guess: text}
        if not isinstance(guess, str):
            guess = None
    except Exception as e:
        guess = None
    try:
        la_id, _ = langid.classify(text)
        det = detect(text)
        if det and la_id and det == la_id and det in probables:
            return {det: text}
    except Exception as e:
        det = None
    return {"XX": text}


def _get_address(g: Graph, iri: BNode):
    """Get address from graph"""
    locality = str(next(g.objects(iri, SDO.addressLocality)))
    code = str(next(g.objects(iri, SDO.postalCode)))
    street = str(next(g.objects(iri, SDO.streetAddress)))
    country = _get_country(locality)
    return {'__type': 'Address',
            'street': street,
            'postalCode': code,
            'locality': locality,
            'country': country}


def _get_country(locality):
    """Get country of an address

    Checks if the place is in Switzerland.
    """
    q = f"http://api.geonames.org/searchJSON?username=blandolt&name_equals={locality}&country=ch"
    r = requests.get(q)
    js = r.json()
    gn = js.get('geonames')
    return "Switzerland" if len(gn) > 0 else "XX - Switzerland?"


def _get_place(g: Graph, iri: BNode):
    """Get a place from graph"""
    url = next(g.objects(iri, SDO.url))
    return _get_url(g, url)


def _ensure_protocol_in_url(url: str):
    """Tries to ensure that a url starts with a valid protocol."""
    if url.startswith('https://') or url.startswith('http://'):
        return url
    else:
        return f'http://{url}'


def _get_url(g: Graph, iri: BNode):
    """Get URL from graph"""
    url = str(next(g.objects(iri, SDO.url))).strip()
    try:
        propID_bnode = next(g.objects(iri, SDO.propertyID))
        propID = str(next(g.objects(propID_bnode, SDO.propertyID)))
    except StopIteration:
        propID = url
    type_ = _get_url_type(propID)
    txt = _get_url_text(url, type_)
    url = _ensure_protocol_in_url(url)
    res = {"__type": "URL",
           "type": type_,
           "url": url
           }
    if txt:
        res['text'] = txt
    return res


def _get_predefined_multilanguage_string(text: str):
    knowns: Dict[str, Dict] = {'German': {'de': 'Deutsch',
                                          'en': 'German',
                                          'fr': 'allemand'},
                               'English': {'en': 'English',
                                           'de': 'Englisch',
                                           'fr': 'anglais'},
                               'Latin': {'en': 'Latin',
                                         'de': 'Latein',
                                         'fr': 'latin'},
                               'French': {'en': 'French',
                                          'de': 'Französisch',
                                          'fr': 'français'},
                               'Italian': {'en': 'Italian',
                                           'de': 'Italienisch',
                                           'fr': 'italien'}}
    known = knowns.get(text)
    if known:
        return known
    if re.search('^\d{4}( ?- ?\d{4})?$', text):  # date
        return {'de': text,
                'en': text,
                'fr': text}


def _get_url_text(url, t):
    """Get a reasonable display text for a URL of a defined type."""
    if t == 'URL':
        return url
    if t == 'Skos':
        return _get_skos_name(url)
    if t == 'Geonames':
        return _get_geonames_name(url)
    if t == 'Pleiades':
        return f"XX: Pleiades URL: {url}"
    if t == 'ORCID':
        return f"ORCID URL: {url}"
    if t == 'Periodo':
        return _get_periodo_name(url)
    if t == 'GND':
        return f"GND URL: {url}"
    if t == 'VIAF':
        return f"VIAF URL: {url}"
    if t == 'Creative Commons':
        return _get_cc_name(url)
    if t == 'Chronontology':
        return _get_chronontology_name(url)
    f"XX: Unknown Type for URL: {url}"


def _get_periodo_name(url: str):
    """Get display text for a Periodo URL"""
    try:
        r = requests.get(url + '.json')
        data = r.json()
        res = data.get('label')
        return res
    except Exception:
        return f'XX: Periodo URL: {url}'


def _get_chronontology_name(url: str):
    """Get display text for a ChronOntology URL"""
    try:
        id_ = url.rsplit('/', maxsplit=1)[-1]
        r = requests.get(f'https://chronontology.dainst.org/data/period/{id_}')
        data = r.json()
        names = data.get('resource').get('names').get('en')
        return names[0]
    except Exception:
        return f'XX: Chronontology URL: {url}'


# LATER: this should not be called "SKOS"
def _get_skos_name(url: str):
    """Get display text for a SKOS URL"""
    url = url.removesuffix('/')
    url = url.removesuffix('/html')
    try:
        url += '/n-triples'
        r = requests.get(url)
        data = r.text
        g = Graph()
        g.parse(data=data, format='n3')
        labels = g.objects(predicate=SKOS.prefLabel)
        for label in labels:
            l: Literal = label
            if l.language == 'en':
                return str(l)
    except Exception:
        return f'XX: Skos URL: {url}'


def _get_cc_name(url: str):
    """Get display text for a Creative Commons URL"""
    try:
        r = requests.get(url.strip())
        soup = BeautifulSoup(r.content, 'html.parser')
        ident = soup.select_one('span.cc-license-identifier')
        res: str = ident.get_text()
        res = res.strip().removeprefix('(').removesuffix(')')
        return res
    except Exception:
        return f'XX: CC URL: {url}'


def _get_geonames_name(url: str):
    """Get display text for a geonames URL"""
    if url.endswith('/'):
        url = url.removesuffix('/')
    try:
        if re.search('\/countries\/', url):
            r = requests.get(url)
            soup = BeautifulSoup(r.content, 'html.parser')
            tables = soup.select('table')
            table = tables[1]
            link = table.select_one('a')
            url = link['href']
        if url.endswith(".html"):
            url = url.rsplit("/", 1)[0]
        gn_id = url.rsplit("/")[-1]
        base = f'http://api.geonames.org/getJSON'
        payload = {'geonameId': gn_id,
                   'username': 'blandolt'}
        r = requests.get(base, params=payload)
        resp = r.json()
        name = resp.get('name')
        if not name:
            raise Exception()
        return name
    except Exception:
        return f'XX: Geonames URL: {url}'


def _get_url_type(propID):
    """Get URL type from schema.propertyID"""
    if propID.startswith("SKOS"):
        return "Skos"
    if propID.startswith("Geonames"):
        return "Geonames"
    if propID.startswith("Pleiades"):
        return "Pleiades"
    if propID.startswith("ORCID"):
        return "ORCID"
    if propID.startswith("Periodo"):
        return "Periodo"
    if propID.startswith("ChronOntology") or propID.startswith("dainst."):
        return "Chronontology"
    if propID.startswith("GND"):
        return "GND"
    if propID.startswith("VIAF"):
        return "VIAF"
    if propID.startswith("Creative Commons"):
        return "Creative Commons"
    else:
        return "URL"


def validate(data, verbose=False):
    """Validates JSON

    Validates json serialized data against the schema of the API specification.  

    [Schema](https://raw.githubusercontent.com/dasch-swiss/dasch-service-platform/main/docs/services/metadata/schema-metadata.json)

    Args:
        data ([type]): [description]
    """
    r = requests.get(schema_url)
    schema = r.json()
    try:
        if verbose:
            print("Validating...")
        validator = jsonschema.Draft7Validator(schema)
        valid = True
        for e in validator.iter_errors(data):
            if verbose:
                print(f'Validation Error: {e.message}')
            valid = False
        if valid and verbose:
            print("JSON is valid.")
        return valid
    except jsonschema.ValidationError as val:
        if verbose:
            print(val.message)
        return False


if __name__ == "__main__":
    files = ['maximal.ttl',
             'rosetta.ttl',
             'limc.ttl',
             'awg.ttl',
             'hdm.ttl',
             'drawings-gods.ttl'
             ]
    results = {}
    os.makedirs('out', exist_ok=True)
    for filename in files:
        print(f'Converting: {filename}...')
        s = convert_file(f'test/test-data/{filename}')
        issues = s.count("XX")
        results[filename] = {
            'isValid': validate(s),
            'numberOfIssues': issues,
            'toResolve': [l.strip().replace('"', "'") for l in s.splitlines() if 'XX' in l]
        }
        print(f'Number of issues encountered: {issues}')
        out = filename.replace('.ttl', '.json')
        with open(f'out/{out}', mode='w+', encoding='utf-8') as f:
            f.write(s)
        print(f'Saved as: {out}\nDone.\n\n----\n')
    print(json.dumps(results, indent=4))
